'use strict';

const cp = require('child_process');
const path = require('path');
const colors = require('colors');
const Package = require('@foundation-cli/package');
const log = require('@foundation-cli/log');

// 命令名称与npm包名称的映射表
const SETTINGS = {
  init: '@foundation-cli/init',
  'offline-publish': '@foundation-cli/offline-publish'
};

const CACHE_DIR = 'dependencies';

async function exec() {
  let targetModule = process.env.CLI_TARGET_MODULE;
  let targetPath = process.env.CLI_TARGET_PATH;
  let storePath = '';
  const homePath = process.env.CLI_HOME_PATH;

  log.verbose('homePath:', homePath);

  const args = Array.from(arguments); // 为什么使用arguments来接收参数，因为exec对于所有命令来说都是通用的，但是每个命令接收的参数个数可能不同
  // const cmdOptions = args[args.length - 2]; // 倒数第二个参数是一个对象，是当前命令的的option的集合
  const cmdObj = args[args.length - 1]; // 倒数第一个参数是当前command对象实例
  const cdmName = cmdObj.name(); // 获取命令的名称
  const packageName = targetModule || SETTINGS[cdmName]; // npm包的名称
  const packageVersion = 'latest'; // npm包的名称

  let pkg;

  if (!targetPath) { // 用户没有指定targetPath，这时候就会去加载本地的缓存
    targetPath = path.resolve(homePath, CACHE_DIR);
    storePath = path.resolve(targetPath, 'node_modules');

    log.verbose('targetPath:', targetPath);
    log.verbose('storePath:', storePath);

    pkg = new Package({
      targetPath,
      storePath,
      packageName,
      packageVersion
    });
    if (await pkg.exists()) { // package存在于缓存中
      // 更新package
      await pkg.update();
    } else { // 当package不存在于缓存中，从远程拉取模块并保存到缓存中
      await pkg.install();
    }
  } else { // 命令行中用户指定了targetPath
    pkg = new Package({
      targetPath,
      storePath,
      packageName,
      packageVersion
    });
  }

  const rootFile = pkg.getRootFilePath();

  if (rootFile) {
    try {
      // require(rootFile).call(null, Array.from(args)); 在当前进程执行命令

      // 在子进程中进行命令执行

      // 对cmdObj（args[args.length - 1]）不能直接传，要进行瘦身
      const  o = Object.create(null);

      Object.keys(cmdObj).forEach(key => {
        // 不拷贝原型链上的属性, 不拷贝_开头的属性，去除parent属性
        if (cmdObj.hasOwnProperty(key) && !key.startsWith('_') && key !== 'parent') {
          o[key] = cmdObj[key];
        }
      });

      args[args.length - 1] = o;

      const code = `require('${rootFile}').call(null, ${JSON.stringify(args)})`;

      // -e是执行代码的意思
      const child = cp.spawn('node', ['-e', code], {
        cwd: process.cwd(),
        // stdio属性很重要，默认是pipe，这时候子进程和父进程之间会有一个管道用来传输数据，这时候要使用下面监听data事件的方式，才能获取到执行过程中的相关信息
        // 设置成inherit以后，相当于把child.stdout，child.stderr直接跟父进程的, process.stderr绑定到一起了，就不需要监听data事件了，会直接打出执行过程中的信息
        stdio: 'inherit'
      });

      /*child.stdout.on('data', (chunk) => {
        console.log(chunk.toString());
      });
      child.stderr.on('data', (chunk) => {
        console.log(chunk.toString());
      });*/

      child.on('error', e => { // 命令执行失败的事件处理
        log.error(colors.red(e.message));
        process.exit(1);
      });

      child.on('exit', e => { // 命令执行结束的事件处理
        log.verbose('命令执行成功：' + e);
        process.exit(e);
      });

    } catch (e) {
      log.error(colors.red(e.message));
    }
  }
}

module.exports = exec;
